from flask import current_app
from jackdaw.dbmodel.aduser import JackDawADUser
from jackdaw.dbmodel.adcomp import JackDawADMachine
from jackdaw.dbmodel.adinfo import JackDawADInfo
from jackdaw.dbmodel.smbfinger import SMBFinger
import base64


class Anomalies:
	def __init__(self, db_conn = None, db_session = None):
		self.db_conn = db_conn
		self.db_session = db_session


	def get_user_pwnotreq(self, domainid, page, maxcnt):
		res = {
			'res' : [],
			'page': {},
		}
		pw_notreq_users = []

		qry = self.db_session.session.query(
			JackDawADUser
			).filter_by(ad_id = domainid
			).filter(JackDawADUser.UAC_PASSWD_NOTREQD == True
			).with_entities(JackDawADUser.id, JackDawADUser.sAMAccountName
			)
			
		qry = qry.paginate(page = page, max_per_page = maxcnt)
		for uid, username in qry.items:
			pw_notreq_users.append(
				{
					'uid' : uid, 
					'username' : username
				}
			)
		page = dict(
			total=qry.total, 
			current_page=qry.page,
			per_page=qry.per_page
		)

		res['res'] = pw_notreq_users
		res['page'] = page

		return res

	def get_user_plaintext(self, domainid, page, maxcnt):
		res = {
			'res' : [],
			'page': {},
		}
		plaintext_pw_users = []
		qry = self.db_session.session.query(
			JackDawADUser
			).filter_by(ad_id = domainid
			).filter(JackDawADUser.UAC_ENCRYPTED_TEXT_PASSWORD_ALLOWED == True
			).with_entities(JackDawADUser.id, JackDawADUser.sAMAccountName
			)

		qry = qry.paginate(page = page, max_per_page = maxcnt)
		
		for uid, username in qry.items:
			plaintext_pw_users.append(
				{
					'uid' : uid, 
					'username' : username
				}
			)

		page = dict(
			total=qry.total, 
			current_page=qry.page,
			per_page=qry.per_page
		)

		res['res'] = plaintext_pw_users
		res['page'] = page

		return res

	def get_user_pw_notexp(self, domainid, page, maxcnt):
		res = {
			'res' : [],
			'page': {},
		}
		pw_notexp = []
		qry = self.db_session.session.query(
			JackDawADUser
			).filter_by(ad_id = domainid
			).filter(JackDawADUser.UAC_DONT_EXPIRE_PASSWD == True
			).with_entities(JackDawADUser.id, JackDawADUser.sAMAccountName
			)

		qry = qry.paginate(page = page, max_per_page = maxcnt)
		for uid, username in qry.items:
			pw_notexp.append(
				{
					'uid' : uid, 
					'username' : username
				}
			)

		page = dict(
			total=qry.total, 
			current_page=qry.page,
			per_page=qry.per_page
		)

		res['res'] = pw_notexp
		res['page'] = page

		return res

	def get_user_des_only(self, domainid, page, maxcnt):
		res = {
			'res' : [],
			'page': {},
		}
		des_only = []
		qry = self.db_session.session.query(
			JackDawADUser
			).filter_by(ad_id = domainid
			).filter(JackDawADUser.UAC_USE_DES_KEY_ONLY == True
			).with_entities(JackDawADUser.id, JackDawADUser.sAMAccountName
			)

		qry = qry.paginate(page = page, max_per_page = maxcnt)
		for uid, username in qry.items:
			des_only.append(
				{
					'uid' : uid, 
					'username' : username
				}
			)

		page = dict(
			total=qry.total, 
			current_page=qry.page,
			per_page=qry.per_page
		)

		res['res'] = des_only
		res['page'] = page

		return res

	def get_user_asrep(self, domainid, page, maxcnt):
		res = {
			'res' : [],
			'page': {},
		}
		asrep = []
		qry = self.db_session.session.query(
			JackDawADUser
			).filter_by(ad_id = domainid
			).filter(JackDawADUser.UAC_DONT_REQUIRE_PREAUTH == True
			).with_entities(JackDawADUser.id, JackDawADUser.sAMAccountName
			)

		qry = qry.paginate(page = page, max_per_page = maxcnt)
		for uid, username in qry.items:
			asrep.append(
				{
					'uid' : uid, 
					'username' : username
				}
			)

		page = dict(
			total=qry.total, 
			current_page=qry.page,
			per_page=qry.per_page
		)

		res['res'] = asrep
		res['page'] = page

		return res

		
	def get_user_description(self, domainid, page, maxcnt):
		res = {
			'res' : [],
			'page': {},
		}
		users = []
		qry = self.db_session.session.query(
			JackDawADUser.id, JackDawADUser.sAMAccountName, JackDawADUser.description
			).filter(JackDawADUser.ad_id == domainid
			).filter(JackDawADUser.description != None
			).filter(JackDawADUser.description != ""
			)

		qry = qry.paginate(page = page, max_per_page = maxcnt)
		for mid, name, description in qry.items:
			users.append({
					'userid' : mid,
					'username' : name,
					'description' : description,
				})
		
		page = dict(
			total=qry.total, 
			current_page=qry.page,
			per_page=qry.per_page
		)

		res['res'] = users
		res['page'] = page

		return res

	def get_machine_description(self, domainid, page, maxcnt):
		res = {
			'res' : [],
			'page': {},
		}
		machines = []
		qry = self.db_session.session.query(
			JackDawADMachine.id, JackDawADMachine.sAMAccountName, JackDawADMachine.description
			).filter(JackDawADMachine.ad_id == domainid
			).filter(JackDawADMachine.description != None
			).filter(JackDawADMachine.description != ""
			)

		qry = qry.paginate(page = page, max_per_page = maxcnt)
		for mid, name, description in qry.items:
			machines.append({
					'machineid' : mid,
					'machinename' : name,
					'description' : description,
				})
		
		page = dict(
			total=qry.total, 
			current_page=qry.page,
			per_page=qry.per_page
		)

		res['res'] = machines
		res['page'] = page

		return res

	def get_machine_outdated(self, domainid, version, page, maxcnt):
		res = {
			'res' : [],
			'page': {},
		}
		version = base64.b64decode(version).decode()
		# TODO: input filtering more? not sure if needed here...
		machines = []
		qry = self.db_session.session.query(
			JackDawADMachine
			).filter_by(ad_id = domainid
			).filter(JackDawADMachine.operatingSystemVersion == version
			).with_entities(JackDawADMachine.id, JackDawADMachine.sAMAccountName, JackDawADMachine.objectSid
			)
		
		qry = qry.paginate(page = page, max_per_page = maxcnt)
		for mid, name, sid in qry.items:
			machines.append({
					'machineid' : mid,
					'machinename' : name,
					'machinesid' : sid
				})
		
		page = dict(
			total=qry.total, 
			current_page=qry.page,
			per_page=qry.per_page
		)

		res['res'] = machines
		res['page'] = page

		return res


	def get_smb_nosig(self, domainid, page, maxcnt):
		res = {
			'res' : [],
			'page': {},
		}
		machines = []
		qry = self.db_session.session.query(
			JackDawADMachine.id, JackDawADMachine.sAMAccountName, SMBFinger.id
			).filter(JackDawADMachine.ad_id == domainid
			).filter(SMBFinger.machine_id == JackDawADMachine.id
			).filter(SMBFinger.signing_required == False
			)

		qry = qry.paginate(page = page, max_per_page = maxcnt)
		for mid, name, _ in qry.items:
			machines.append({
					'id' : mid,
					'name' : name,
				})
		
		page = dict(
			total=qry.total, 
			current_page=qry.page,
			per_page=qry.per_page
		)

		res['res'] = machines
		res['page'] = page

		return res


	def get_smb_domain_mismatch(self, domainid, page, maxcnt):
		res = {
			'res' : [],
			'page': {},
		}
		machines = []
		qry = self.db_session.session.query(
			JackDawADInfo.id, JackDawADMachine.id, JackDawADMachine.sAMAccountName , SMBFinger.domainname
			).filter(JackDawADMachine.ad_id == domainid
			).filter(JackDawADInfo.id == domainid
			).filter(SMBFinger.machine_id == JackDawADMachine.id
			).filter(SMBFinger.domainname != JackDawADInfo.name
			).filter(SMBFinger.domainname != None
			)

		qry = qry.paginate(page = page, max_per_page = maxcnt)
		for _, mid, name, domain in qry.items:
			machines.append({
					'machineid' : mid,
					'machinename' : name,
					'domainname' : domain,
				})
		
		page = dict(
			total=qry.total, 
			current_page=qry.page,
			per_page=qry.per_page
		)

		res['res'] = machines
		res['page'] = page

		return res










class GraphConstruct:
	def __init__(self, ad_id, diff_name = None):
		self.ad_id = ad_id
		self.diff_name = diff_name # should be a string 'new' or 'old'
		
		self.include_nodes = []
		self.include_edges = []

		self.blacklist_sids = {'S-1-5-32-545': ''}
		self.ignoresids = {"S-1-3-0": '', "S-1-5-18": ''}


	def is_blacklisted_sid(self, sid):
		if sid in self.blacklist_sids:
			return True
		if sid[:len('S-1-5-21')] == 'S-1-5-21':
			if sid[-3:] == '513':
				return True
			
		return False


from flask import current_app
from jackdaw.dbmodel.aduser import JackDawADUser
from jackdaw.dbmodel.adcomp import JackDawADMachine
from jackdaw.dbmodel.adinfo import JackDawADInfo
from jackdaw.dbmodel.smbfinger import SMBFinger
from jackdaw.nest.anomalies.anomalies import Anomalies


def get_user_pwnotreq(domainid, page, maxcnt):
	anom = Anomalies(db_session = current_app.db)
	res = anom.get_user_pwnotreq(domainid, page, maxcnt)
	return res

def get_user_plaintext(domainid, page, maxcnt):
	anom = Anomalies(db_session = current_app.db)
	res = anom.get_user_plaintext(domainid, page, maxcnt)
	return res

def get_user_pw_notexp(domainid, page, maxcnt):
	anom = Anomalies(db_session = current_app.db)
	res = anom.get_user_pw_notexp(domainid, page, maxcnt)
	return res

def get_user_des_only(domainid, page, maxcnt):
	anom = Anomalies(db_session = current_app.db)
	res = anom.get_user_des_only(domainid, page, maxcnt)
	return res

def get_user_asrep(domainid, page, maxcnt):
	anom = Anomalies(db_session = current_app.db)
	res = anom.get_user_asrep(domainid, page, maxcnt)
	return res

def get_user_description(domainid, page, maxcnt):
	anom = Anomalies(db_session = current_app.db)
	res = anom.get_user_description(domainid, page, maxcnt)
	return res

def get_machine_description(domainid, page, maxcnt):
	anom = Anomalies(db_session = current_app.db)
	res = anom.get_machine_description(domainid, page, maxcnt)
	return res

def get_machine_outdated(domainid, version, page, maxcnt):
	anom = Anomalies(db_session = current_app.db)
	res = anom.get_machine_outdated(domainid, version, page, maxcnt)
	return res

def get_smb_nosig(domainid, page, maxcnt):
	anom = Anomalies(db_session = current_app.db)
	res = anom.get_smb_nosig(domainid, page, maxcnt)
	return res

def get_smb_domain_mismatch(domainid, page, maxcnt):
	anom = Anomalies(db_session = current_app.db)
	res = anom.get_smb_domain_mismatch(domainid, page, maxcnt)
	return res






from flask_sqlalchemy import SQLAlchemy
from jackdaw.dbmodel.adinfo import JackDawADInfo
from flask import current_app

def list_domains(page, maxcnt):
	db = current_app.db
	res = {
		'res' : [],
		'page': {},
	}
	qry = db.session.query(
		JackDawADInfo
		).with_entities(
			JackDawADInfo.id, 
			JackDawADInfo.distinguishedName, 
			JackDawADInfo.fetched_at
			)
		
	qry = qry.paginate(page = page, max_per_page = maxcnt)
	domains = []
	for did, distinguishedName, creation in qry.items:
		name = distinguishedName.replace('DC=','')
		name = name.replace(',','.')
		domains.append(
			{
				'id': did,
				'name': name, 
				'creation' : creation
			}
		)

	page = dict(
		total=qry.total, 
		current_page=qry.page,
		per_page=qry.per_page
	)

	res['res'] = domains
	res['page'] = page

	return res

def get(domainid):
	db = current_app.db
	adinfo = db.session.query(JackDawADInfo).get(domainid)
	return adinfo.to_dict()




from jackdaw.dbmodel.netsession import NetSession
from jackdaw.dbmodel.adcomp import JackDawADMachine
from jackdaw.dbmodel.aduser import JackDawADUser
from flask import current_app

def session_list(domainid):
	db = current_app.db
	sessions = {}
	for mid, mname, session in db.session.query(JackDawADMachine.id, JackDawADMachine.sAMAccountName, NetSession).filter(JackDawADMachine.ad_id == domainid).filter(NetSession.machine_id == JackDawADMachine.id).distinct(NetSession.username):
		if mid not in sessions:
			sessions[mid] = {}
			sessions[mid]['sessions'] = []

		sessions[mid]['machinename'] = mname
		sessions[mid]['sessions'].append(session.username)

	return sessions

def session_add(domainid, session):
	db = current_app.db
	print(session)
	cname = session['hostname']
	if cname[-1] != '$':
		cname = session['hostname'] + '$'
	comp = db.session.query(JackDawADMachine.id, JackDawADMachine.sAMAccountName).filter_by(ad_id = domainid).filter(JackDawADMachine.sAMAccountName == cname).first()
	if comp is None:
		return 'Machine not found!', 404
	uname = session['username']
	user = db.session.query(JackDawADUser.sAMAccountName).filter_by(ad_id = domainid).filter(JackDawADUser.sAMAccountName == uname).first()
	if user is None:
		return 'User not found!', 404

	sess = NetSession()
	sess.machine_id = comp.id
	sess.source = comp.sAMAccountName
	sess.username = user.sAMAccountName
	try:
		db.session.add(sess)
		db.session.commit()
	except:
		db.session.rollback()

	return 'Session created!', 200




from jackdaw.dbmodel.adinfo import JackDawADInfo
from jackdaw.dbmodel.tokengroup import JackDawTokenGroup
from jackdaw.dbmodel import *
from jackdaw.wintypes.well_known_sids import get_name_or_sid, get_sid_for_name
from jackdaw.wintypes.lookup_tables import *
from sqlalchemy import not_, and_, or_, case
from sqlalchemy.orm import load_only
#from pyvis.network import Network
#from pyvis.options import Layout
import networkx as nx

from jackdaw.nest.graph.graphdata import *
from jackdaw import logger
import enum


def ace_applies(ace_guid, object_class):
	'''
	Checks if an ACE applies to this object (based on object classes).
	Note that this function assumes you already verified that InheritedObjectType is set (via the flag).
	If this is not set, the ACE applies to all object types.
	'''
	try:
		our_ace_guid = OBJECTTYPE_GUID_MAP[object_class]
	except KeyError:
		return False
	if ace_guid == our_ace_guid:
		return True
	# If none of these match, the ACE does not apply to this object
	return False

class NodeType(enum.Enum):
	USER = 'USER'
	GROUP = 'GROUP'
	MACHINE = 'MACHINE'
	
class EdgeType(enum.Enum):
	MEMBER_GROUP = 'MEMBER_GROUP'
	MEMBER_USER = 'MEMBER_USER'
	SESSION = 'SESSION'
	ACL = 'ACL'
	LOCALGROUP = 'LOCALGROUP'
	PWSHARE = 'PWSHARE'
	CUSTOM = 'CUSTOM'
	DELEGATION_CONSTRAINED = 'DELEGATION_CONSTRAINED'
	DELEGATION_UNCONSTRAINED = 'DELEGATION_UNCONSTRAINED'


class DomainGraph:
	def __init__(self, db_conn = None, dbsession = None):
		self.db_conn = db_conn
		self.dbsession = dbsession
		self.constructs = {}
		self.graph = nx.DiGraph()

		
	def get_session(self):
		if self.db_conn is not None:
			return get_session(self.db_conn)
		elif self.dbsession is not None:
			return self.dbsession
		else:
			raise Exception('Either db_conn or dbsession MUST be supplied!')

	######################################################################################################################################
	#################################################      PATH CALCULATION PART         #################################################
	######################################################################################################################################
			
	def sid2cn(self, sid, throw = False):
		session = self.get_session()
		tsid = session.query(JackDawTokenGroup.cn).filter(JackDawTokenGroup.sid == sid).first()
		#print('sid2cn: %s' % tsid)
		if not tsid:
			t = str(get_name_or_sid(str(sid)))
			if t == sid and throw == True:
				raise Exception('No CN found for SID = %s' % repr(sid))
			return t
		return tsid[0]
	
	def cn2sid(self, cn, throw = False, domain_sid = None):
		sid = get_sid_for_name(cn, domain_sid)
		
		session = self.get_session()
		tsid = session.query(JackDawTokenGroup.sid).filter(JackDawTokenGroup.cn == cn).first()
		#print(tsid)
		if not tsid:
			if throw == True:
				raise Exception('No SID found for CN = %s' % repr(cn))
			return cn
		return tsid[0]
			
	def show_all(self):
		"""
		Returns all nodes and edges from the graph
		You really don't want this on a larger graph, mostly used for testing purposes
		"""
		data = GraphData()
		for sid in self.graph.nodes:
			data.add_node(
				sid, 
				self.graph.nodes[sid].get('name', self.sid2cn(sid)), 
				self.graph.nodes[sid]['construct'].ad_id,
				self.graph.nodes[sid].get('node_type')
			)

		for edge in [e for e in self.graph.edges]:
			data.add_edge(edge[0], edge[1])

		return data
	
	def get_node(self, nodeid = None):
		
		if nodeid is None:
			nodes = []
			for gid, props in list(self.graph.nodes(data=True)):
				nodes.append(GraphNode(gid, gid, props['construct'].ad_id )) #properties = {}))
			return nodes

		if nodeid not in self.graph.nodes:
			return None
		gid, props = self.graph.nodes[nodeid]
		return GraphNode(gid, gid, props['construct'].ad_id, properties = {})
			
	def __add_path(self, network, path):
		"""
		Adds the path to the representational network
		Path is a list of sids (nodes), so we need to find the edges matching
		"""
		#print('PATH: %s' % repr(path))
		for d, sid in enumerate(path):
			#print(sid)
			network.add_node(
				sid, 
				name = self.graph.nodes[sid].get('name', self.sid2cn(sid)), 
				node_type = self.graph.nodes[sid].get('node_type'),
				domainid = self.graph.nodes[sid]['construct'].ad_id
			)
			network.nodes[sid].set_distance(d)
		
		for i in range(len(path) - 1):
			for edge in self.graph.edges([path[i], ], data=True):
				if edge[1] == path[i + 1]:
					name = edge[2].get('label', None)
					network.add_edge(path[i], path[i + 1], label=name)

	def distances_from_node(self, dst_sid):
		
		distances = {} #distance -> occurrence of distance
		for node in self.graph.nodes:
			try:
				for path in nx.all_shortest_paths(self.graph, source = node, target = dst_sid):
					distance = len(path)
					if distance not in distances:
						distances[distance] = 0
					distances[distance] += 1
			except nx.exception.NetworkXNoPath:
				continue

		return distances

	def all_shortest_paths(self, src_sid = None, dst_sid = None):
		nv = GraphData()
		
		if not src_sid and not dst_sid:
			raise Exception('Either source or destination MUST be specified')
		
		elif not src_sid and dst_sid:
			#for each node we calculate the shortest path to the destination node, silently skip the ones who do not have path to dst
			for node in self.graph.nodes:
				if node == dst_sid:
					continue
				try:
					for path in nx.all_shortest_paths(self.graph, source = node, target = dst_sid):
						self.__add_path(nv, path)
									
				except nx.exception.NetworkXNoPath:
					continue
			
		elif src_sid and not dst_sid:
			#for each node we calculate the shortest path to the destination node, silently skip the ones who do not have path to dst
			
			for node in self.graph.nodes:
				if node == src_sid:
					continue
				try:
					for path in nx.all_shortest_paths(self.graph, source = src_sid, target = node):
						self.__add_path(nv, path)
									
				except nx.exception.NetworkXNoPath:
					continue
					
		else:
			#for each node we calculate the shortest path to the destination node, silently skip the ones who do not have path to dst
			for path in nx.all_shortest_paths(self.graph, source = src_sid, target = dst_sid):
				self.__add_path(nv, path)
		
		return nv
	

	######################################################################################################################################
	#################################################      GRAPH CALCULATION PART        #################################################
	######################################################################################################################################

	def add_sid_to_node(self, node, node_type, construct, name = None):
		if construct.is_blacklisted_sid(node):
			return
		if not name:
			name = str(get_name_or_sid(str(node)))
		
		#this presence-filter is important, as we will encounter nodes that are known and present in the graph already
		#ald later will be added via tokengroups as unknown
		if node not in self.graph.nodes:
			self.graph.add_node(str(node), name=name, node_type=node_type, construct = construct)
		
			
	def add_edge(self, sid_src, sid_dst, construct, label = None, weight = 1):
		if not sid_src or not sid_dst:
			return
		if construct.is_blacklisted_sid(sid_src) or construct.is_blacklisted_sid(sid_dst):
			return
			
		self.add_sid_to_node(sid_src, 'unknown', construct)
		self.add_sid_to_node(sid_dst, 'unknown', construct)
			
		self.graph.add_edge(sid_src, sid_dst, label = label, weight = weight)
		
		#self.network_visual.add_edge(sid_src, sid_dst, label = label, weight = weight)
	
	def calc_acl_edges(self, adinfo, construct):
		for acl in adinfo.objectacls:
			##SUPER-IMPORTANT!!!!
			##THE ACCESS RIGHTS CALCULATIONS ARE FLAWED (JUST LIKE IN BLOODHOUND)
			##CORRECT WAY WOULD BE TO TRAVERSE ALL ACES IN A GIVEN ACL IN THE ORDER OF "ACE-ODER" AND CHECK FOR DENY POLICIES AS WELL!!!
			##TODO: CALCULATE THE EFFECTIVE PERMISSIONS IN A CORRECT MANNER :)
			
			##TODO: The DB is designed that most of these calculations can (and should) be offloaded to the DB itself
			##      Currently it's in python bc I'm laze and Dirkjan already wrote the major part
			##      https://github.com/fox-it/BloodHound.py/blob/0d3897ba4cc00818c0fc3e01fa3a91fc42e799e2/bloodhound/enumeration/acls.py
			
			if acl.owner_sid not in construct.ignoresids:
				self.add_edge(acl.owner_sid, acl.sid, construct, label='Owner')
				
			if acl.ace_sid in construct.ignoresids:
				continue
			
			if acl.ace_type in ['ACCESS_ALLOWED_ACE_TYPE','ACCESS_ALLOWED_OBJECT_ACE_TYPE']:
				if acl.ace_type == 'ACCESS_ALLOWED_ACE_TYPE':
					if acl.ace_mask_generic_all == True:
						self.add_edge(acl.ace_sid, acl.sid, construct, label='GenericALL')
					
					if acl.ace_mask_generic_write == True:
						self.add_edge(acl.ace_sid, acl.sid, construct, label='GenericWrite')
						
					if acl.ace_mask_write_owner == True:
						self.add_edge(acl.ace_sid, acl.sid, construct, label='WriteOwner')
						
					if acl.ace_mask_write_dacl == True:
						self.add_edge(acl.ace_sid, acl.sid, construct, label='WriteDacl')
						
					if acl.object_type in ['user', 'domain'] and acl.ace_mask_control_access == True:
						self.add_edge(acl.ace_sid, acl.sid, construct, label='ExtendedRightALL')
				
				if acl.ace_type == 'ACCESS_ALLOWED_OBJECT_ACE_TYPE':
					if acl.ace_hdr_flag_inherited == True and acl.ace_hdr_flag_inherit_only == True:
						continue
					
					if acl.ace_hdr_flag_inherited == True and acl.ace_inheritedobjecttype is not None:
						if not ace_applies(acl.ace_inheritedobjecttype, acl.object_type):
							continue
					
					if any([acl.ace_mask_generic_all, acl.ace_mask_write_dacl, acl.ace_mask_write_owner, acl.ace_mask_generic_write]):
						if acl.ace_objecttype is not None and not ace_applies(acl.ace_objecttype, acl.object_type):
							continue
						
						if acl.ace_mask_generic_all == True:
							self.add_edge(acl.ace_sid, acl.sid, construct, label='GenericALL')
							continue
				
						if acl.ace_mask_generic_write == True:
							self.add_edge(acl.ace_sid, acl.sid, construct, label='GenericWrite')
							
							if acl.object_type != 'domain':
								continue
							
						if acl.ace_mask_write_dacl == True:
							self.add_edge(acl.ace_sid, acl.sid, construct, label='WriteDacl')
							
						if acl.ace_mask_write_owner == True:
							self.add_edge(acl.ace_sid, acl.sid, construct, label='WriteOwner')
							
					if acl.ace_mask_write_prop == True:
						if acl.object_type in ['user','group'] and acl.ace_objecttype is None:
							self.add_edge(acl.ace_sid, acl.sid, construct, label='GenericWrite')
							
						if acl.object_type == 'group' and acl.ace_objecttype == 'bf9679c0-0de6-11d0-a285-00aa003049e2':
							self.add_edge(acl.ace_sid, acl.sid, construct, label='AddMember')
							
						
				
					if acl.ace_mask_control_access == True:
						if acl.object_type in ['user','group'] and acl.ace_objecttype is None:
							self.add_edge(acl.ace_sid, acl.sid, construct, label='ExtendedAll')
						
						if acl.object_type == 'domain' and acl.ace_objecttype == '1131f6ad-9c07-11d1-f79f-00c04fc2dcd2':
							# 'Replicating Directory Changes All'
							self.add_edge(acl.ace_sid, acl.sid, construct, label='GetChangesALL')
								
						if acl.object_type == 'domain' and acl.ace_objecttype == '1131f6aa-9c07-11d1-f79f-00c04fc2dcd2':
								# 'Replicating Directory Changes'
								self.add_edge(acl.ace_sid, acl.sid, construct, label='GetChanges')
								
						if acl.object_type == 'user' and acl.ace_objecttype == '00299570-246d-11d0-a768-00aa006e0529':
								# 'Replicating Directory Changes'
								self.add_edge(acl.ace_sid, acl.sid, construct, label='User-Force-Change-Password')
		
	def construct(self, construct):
		"""
		Fills the network graph from database to memory
		"""
		#self.ad_id = ad_id
		session = self.get_session()
		adinfo = session.query(JackDawADInfo).get(construct.ad_id)
		
		self.domain_sid = str(adinfo.objectSid)
	
		#adding group nodes
		for group in adinfo.groups:
			self.add_sid_to_node(group.sid, 'group', construct, name=group.name)
		
		for user in adinfo.users:
			self.add_sid_to_node(user.objectSid, 'user', construct, name=user.sAMAccountName)
				
		#adding user nodes
		for user in adinfo.computers:
			self.add_sid_to_node(user.objectSid, 'machine', construct, name=user.sAMAccountName)
		
		
		for res in session.query(JackDawADUser.objectSid, JackDawADMachine.objectSid).filter(NetSession.username == JackDawADUser.sAMAccountName).filter(NetSession.source == JackDawADMachine.sAMAccountName).distinct(NetSession.username):
			self.add_edge(res[0], res[1], construct, label='hasSession')
			self.add_edge(res[1], res[0], construct, label='hasSession')
		

		for res in session.query(JackDawADUser.objectSid, JackDawADMachine.objectSid, LocalGroup.groupname
					).filter(JackDawADMachine.id == LocalGroup.machine_id
					).filter(JackDawADMachine.ad_id == construct.ad_id
					).filter(JackDawADUser.ad_id == construct.ad_id
					).filter(JackDawADUser.objectSid == LocalGroup.sid
					).all():
			label = None
			if res[2] == 'Remote Desktop Users':
				label = 'canRDP'
				weight = 1
					
			elif res[2] == 'Distributed COM Users':
				label = 'executeDCOM'
				weight = 1
					
			elif res[2] == 'Administrators':
				label = 'adminTo'
				weight = 1
					
			self.add_edge(res[0], res[1], construct, label=label, weight = weight)

		# TODO: implement this!	
		#if self.show_constrained_delegations == True:
		#	pass
			
		# TODO: implement this!	
		#if self.show_unconstrained_delegations == True:
		#	pass

		# TODO: implement this!	
		#for relation in construct.custom_relations:
		#	relation.calc()
		#	self.add_edge(res.sid, res.target_sid)
			
		#print('adding membership edges')
		#adding membership edges
		for tokengroup in adinfo.group_lookups:
			self.add_sid_to_node(tokengroup.sid, 'unknown', construct)
			self.add_sid_to_node(tokengroup.member_sid, 'unknown', construct)
				
			if tokengroup.is_user == True:
				try:
					self.add_edge(tokengroup.sid, tokengroup.member_sid, construct, label='member')
				except AssertionError as e:
					logger.exception()
			elif tokengroup.is_machine == True:
				try:
					self.add_edge(tokengroup.sid, tokengroup.member_sid, construct, label='member')
				except AssertionError as e:
					logger.exception()
			elif tokengroup.is_group == True:
				try:
					self.add_edge(tokengroup.sid, tokengroup.member_sid, construct, label='member')
				except AssertionError as e:
					logger.exception()
		
		logger.info('Adding ACL edges')
		#adding ACL edges
		self.calc_acl_edges(adinfo, construct)


		def get_sid_by_nthash(ad_id, nt_hash):
			return session.query(JackDawADUser.objectSid, Credential.username
				).filter_by(ad_id = ad_id
				).filter(Credential.username == JackDawADUser.sAMAccountName
				).filter(Credential.nt_hash == nt_hash
				)

		dup_nthashes_qry = session.query(Credential.nt_hash
					).filter(Credential.history_no == 0
					).filter(Credential.ad_id == construct.ad_id
                       ).filter(Credential.username != 'NA'
                       ).filter(Credential.domain != '<LOCAL>'
					).group_by(
						Credential.nt_hash
					).having(
						func.count(Credential.nt_hash) > 1
					)

		for res in dup_nthashes_qry.all():
			sidd = {}
			for sid, _ in get_sid_by_nthash(construct.ad_id, res[0]).all():
				sidd[sid] = 1

			for sid1 in sidd:
				for sid2 in sidd:
					if sid1 == sid2:
						continue
					self.add_edge(sid1, sid2, construct, label = 'pwsharing')

		
						
				
					


import traceback
import os
from jackdaw.dbmodel.spnservice import JackDawSPNService
from jackdaw.dbmodel.addacl import JackDawADDACL
from jackdaw.dbmodel.adgroup import JackDawADGroup
from jackdaw.dbmodel.adinfo import JackDawADInfo
from jackdaw.dbmodel.aduser import JackDawADUser
from jackdaw.dbmodel.adcomp import JackDawADMachine
from jackdaw.dbmodel.adou import JackDawADOU
from jackdaw.dbmodel.usergroup import JackDawGroupUser
from jackdaw.dbmodel.adinfo import JackDawADInfo
from jackdaw.dbmodel.tokengroup import JackDawTokenGroup
from jackdaw.dbmodel import *
from jackdaw.wintypes.lookup_tables import *
from jackdaw import logger

from msldap.ldap_objects import *

import multiprocessing
import enum

class LDAPAgentCommand(enum.Enum):
	SPNSERVICE = 0
	SPNSERVICES = 1
	USER = 2
	USERS = 3
	MACHINE = 4
	MACHINES = 5
	OU = 6
	OUS = 7
	DOMAININFO = 8
	GROUP = 9
	GROUPS = 10
	MEMBERSHIP = 11
	MEMBERSHIPS = 12
	SD = 13
	SDS = 14
	EXCEPTION = 99

	SPNSERVICES_FINISHED = 31
	USERS_FINISHED = 32
	MACHINES_FINISHED = 33
	OUS_FINISHED = 34
	GROUPS_FINISHED = 35
	MEMBERSHIPS_FINISHED = 36
	SDS_FINISHED = 37
	DOMAININFO_FINISHED = 38

class LDAPAgentJob:
	def __init__(self, command, data):
		self.command = command
		self.data = data

class LDAPEnumeratorAgent(multiprocessing.Process):
	def __init__(self, ldap_mgr, agent_in_q, agent_out_q):
		multiprocessing.Process.__init__(self)
		self.ldap_mgr = ldap_mgr
		self.agent_in_q = agent_in_q
		self.agent_out_q = agent_out_q
		self.ldap = None

	def get_effective_memberships(self, membership_attr):
		try:
			for sid in self.ldap.get_tokengroups(membership_attr['dn']):
				s = JackDawTokenGroup()
				s.cn = membership_attr['cn']
				s.dn = membership_attr['dn']
				s.guid = membership_attr['guid']
				s.sid = membership_attr['sid']
				s.member_sid = sid
				s.is_user = True if membership_attr['type'] == 'user' else False
				s.is_group = True if membership_attr['type'] == 'group' else False
				s.is_machine = True if membership_attr['type'] == 'machine' else False
				self.agent_out_q.put((LDAPAgentCommand.MEMBERSHIP, s))
		except Exception as e:
			self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
		finally:
			self.agent_out_q.put((LDAPAgentCommand.MEMBERSHIPS_FINISHED, None))

	def enumerate_spnservices(self):
		pass

	def enumerate_machine(self, machine):
		pass
		
	def get_all_spnservices(self):
		try:
			ldap_filter = r'(&(sAMAccountType=805306369))'
			attributes = ['sAMAccountName', 'servicePrincipalName']
			
			for entry in self.ldap.pagedsearch(ldap_filter, attributes):
				for spn in entry['attributes']['servicePrincipalName']:			
					port = None
					service, t = spn.rsplit('/',1)
					m = t.find(':')
					if m != -1:
						computername, port = spn.rsplit(':',1)
					else:
						computername = t
						
					s = JackDawSPNService()
					s.computername = computername
					s.service = service
					s.port = port
					self.agent_out_q.put((LDAPAgentCommand.SPNSERVICE, s))
		except Exception as e:
			self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
		finally:
			self.agent_out_q.put((LDAPAgentCommand.SPNSERVICES_FINISHED, None))

	def get_all_users(self):
		try:
			for user in self.ldap.get_all_user_objects():
				#TODO: fix this ugly stuff here...
				if user.sAMAccountName[-1] == "$":
					continue
				self.agent_out_q.put((LDAPAgentCommand.USER, JackDawADUser.from_aduser(user)))
		except Exception as e:
			self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
		finally:
			self.agent_out_q.put((LDAPAgentCommand.USERS_FINISHED, None))

	def get_all_groups(self):
		try:
			for group in self.ldap.get_all_groups():
				self.agent_out_q.put((LDAPAgentCommand.GROUP, JackDawADGroup.from_dict(group.to_dict())))
		except Exception as e:
			self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
		finally:
			self.agent_out_q.put((LDAPAgentCommand.GROUPS_FINISHED, None))


	def get_all_machines(self):
		try:
			for machine in self.ldap.get_all_machine_objects():
				self.agent_out_q.put((LDAPAgentCommand.MACHINE, JackDawADMachine.from_adcomp(machine)))
		except Exception as e:
			self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
		finally:
			self.agent_out_q.put((LDAPAgentCommand.MACHINES_FINISHED, None))
	
	def get_all_ous(self):
		try:
			for ou in self.ldap.get_all_ous():
				self.agent_out_q.put((LDAPAgentCommand.OU, JackDawADOU.from_adou(ou)))
		except Exception as e:
			self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
		finally:
			self.agent_out_q.put((LDAPAgentCommand.OUS_FINISHED, None))

	def get_domain_info(self):
		try:
			info = self.ldap.get_ad_info()
			adinfo = JackDawADInfo.from_dict(info.to_dict())
			self.agent_out_q.put((LDAPAgentCommand.DOMAININFO, adinfo))
		except Exception as e:
			self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
		finally:
			self.agent_out_q.put((LDAPAgentCommand.DOMAININFO_FINISHED, None))

	def get_sds(self, data):
		try:
			dn = data['dn']
			obj_type = data['obj_type']
			for sd in self.ldap.get_objectacl_by_dn(dn):
				if not sd.nTSecurityDescriptor or not sd.nTSecurityDescriptor.Dacl:
					return
				self.agent_out_q.put((LDAPAgentCommand.SD, {'sd':sd, 'obj_type': obj_type}))
		
		except Exception as e:
			self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
		finally:
			self.agent_out_q.put((LDAPAgentCommand.SDS_FINISHED, None))

	def setup(self):
		try:
			self.ldap = self.ldap_mgr.get_connection()
			self.ldap.connect()
			return True
		except Exception as e:
			self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
			return False

	def run(self):
		res = self.setup()
		#print('agent setup res: %s' % res)
		if res is False:
			return
		while True:
			res = self.agent_in_q.get()
			#print('Got new job! %s' % res)
			if res is None:
				return
			
			if res.command == LDAPAgentCommand.DOMAININFO:
				self.get_domain_info()
			elif res.command == LDAPAgentCommand.USERS:
				self.get_all_users()
			elif res.command == LDAPAgentCommand.MACHINES:
				self.get_all_machines()
			elif res.command == LDAPAgentCommand.GROUPS:
				self.get_all_groups()
			elif res.command == LDAPAgentCommand.OUS:
				self.get_all_ous()
			elif res.command == LDAPAgentCommand.SPNSERVICES:
				self.get_all_spnservices()
			elif res.command == LDAPAgentCommand.MEMBERSHIPS:
				self.get_effective_memberships(res.data)
			elif res.command == LDAPAgentCommand.SDS:
				self.get_sds(res.data)

class LDAPEnumeratorManager:
	def __init__(self, db_conn, ldam_mgr, agent_cnt = 5):
		self.db_conn = db_conn
		self.ldam_mgr = ldam_mgr

		self.session = None

		self.agent_in_q = multiprocessing.Queue()
		self.agent_out_q = multiprocessing.Queue()
		self.agents = []
		self.agent_cnt = agent_cnt
		self.ad_id = None

		self.user_ctr = 0
		self.machine_ctr = 0
		self.ou_ctr = 0
		self.group_ctr = 0
		self.sd_ctr = 0
		self.spn_ctr = 0
		self.member_ctr = 0
		self.domaininfo_ctr = 0

		self.user_finish_ctr = 0
		self.machine_finish_ctr = 0
		self.ou_finish_ctr = 0
		self.group_finish_ctr = 0
		self.sd_finish_ctr = 0
		self.spn_finish_ctr = 0
		self.member_finish_ctr = 0
		self.domaininfo_finish_ctr = 0


	@staticmethod
	def spn_to_account(spn):
		if spn.find('/') != -1:
			return spn.rsplit('/')[1].upper() + '$'
	
	def setup(self):
		logger.info('mgr setup')
		self.session = get_session(self.db_conn)
		for _ in range(self.agent_cnt):
			agent = LDAPEnumeratorAgent(self.ldam_mgr, self.agent_in_q, self.agent_out_q)
			agent.daemon = True
			agent.start()
			self.agents.append(agent)

	def enum_domain(self, info):
		#print('Got domain!')
		self.session.add(info)
		self.session.commit()
		self.session.refresh(info)
		self.ad_id = info.id

		self.user_ctr += 1
		job = LDAPAgentJob(LDAPAgentCommand.USERS, self.ad_id)
		self.agent_in_q.put(job)

		self.machine_ctr += 1
		job = LDAPAgentJob(LDAPAgentCommand.MACHINES, self.ad_id)
		self.agent_in_q.put(job)

		self.group_ctr += 1
		job = LDAPAgentJob(LDAPAgentCommand.GROUPS, self.ad_id)
		self.agent_in_q.put(job)

		self.ou_ctr += 1
		job = LDAPAgentJob(LDAPAgentCommand.OUS, self.ad_id)
		self.agent_in_q.put(job)

		self.spn_ctr += 1
		job = LDAPAgentJob(LDAPAgentCommand.SPNSERVICES, self.ad_id)
		self.agent_in_q.put(job)


	def enum_machine(self, machine):
		#print('Got machine object!')
		machine.ad_id = self.ad_id
		self.session.add(machine)
		self.session.commit()
		self.session.refresh(machine)
		
		for spn in getattr(machine,'allowedtodelegateto',[]):
			con = JackDawMachineConstrainedDelegation()
			con.spn = spn
			con.targetaccount = LDAPEnumeratorManager.spn_to_account(spn)
			machine.allowedtodelegateto.append(con)
		
		self.session.commit()

		membership_attr = {
			'dn'  : str(machine.dn),
			'cn'  : str(machine.cn),
			'guid': str(machine.objectGUID),
			'sid' : str(machine.objectSid),
			'type': 'machine'
		}

		self.member_ctr += 1
		job = LDAPAgentJob(LDAPAgentCommand.MEMBERSHIPS, membership_attr)
		self.agent_in_q.put(job)
		
		self.sd_ctr += 1
		job = LDAPAgentJob(LDAPAgentCommand.SDS, {'dn' : machine.dn, 'obj_type': 'machine' })
		self.agent_in_q.put(job)		

	def enum_user(self, user):
		user.ad_id = self.ad_id
		self.session.add(user)
		self.session.commit()
		self.session.refresh(user)

		for spn in getattr(user,'allowedtodelegateto',[]):
			con = JackDawUserConstrainedDelegation()
			con.spn = spn
			con.targetaccount = LDAPEnumeratorManager.spn_to_account(spn)
			user.allowedtodelegateto.append(con)
		
		self.session.commit()

		membership_attr = {
			'dn'  : str(user.dn),
			'cn'  : str(user.cn),
			'guid': str(user.objectGUID),
			'sid' : str(user.objectSid),
			'type': 'user'
		}

		self.member_ctr += 1
		job = LDAPAgentJob(LDAPAgentCommand.MEMBERSHIPS, membership_attr)
		self.agent_in_q.put(job)

		self.sd_ctr += 1
		job = LDAPAgentJob(LDAPAgentCommand.SDS, {'dn' : user.dn, 'obj_type': 'user' })
		self.agent_in_q.put(job)

	def enum_group(self, group):
		group.ad_id = self.ad_id
		self.session.add(group)
		self.session.commit()
		
		membership_attr = {
			'dn'  : str(group.dn),
			'cn'  : str(group.cn),
			'guid': str(group.guid),
			'sid' : str(group.sid),
			'type': 'group'
		}
		
		self.member_ctr += 1
		job = LDAPAgentJob(LDAPAgentCommand.MEMBERSHIPS, membership_attr)
		self.agent_in_q.put(job)

		self.sd_ctr += 1
		job = LDAPAgentJob(LDAPAgentCommand.SDS, {'dn' : group.dn, 'obj_type': 'group' })
		self.agent_in_q.put(job)

	def enum_ou(self, ou):
		ou.ad_id = self.ad_id
		self.session.add(ou)
		self.session.commit()
		
		self.sd_ctr += 1
		job = LDAPAgentJob(LDAPAgentCommand.SDS, {'dn' : ou.dn, 'obj_type': 'ou' })
		self.agent_in_q.put(job)

	def store_spnservice(self, spn):
		#print('Got SPNSERVICE!')
		spn.ad_id = self.ad_id
		self.session.add(spn)
		self.session.commit()


	def store_membership(self, res):
		#print('Got membership object!')
		res.ad_id = self.ad_id
		self.session.add(res)
		self.session.commit()

	def store_sd(self, data):
		#print('Got SD object!')
		sd = data['sd']
		obj_type = data['obj_type']
		order_ctr = 0
		for ace in sd.nTSecurityDescriptor.Dacl.aces:
			acl = JackDawADDACL()
			acl.ad_id = self.ad_id
			acl.object_type = obj_type
			acl.object_type_guid = OBJECTTYPE_GUID_MAP.get(obj_type)
			acl.owner_sid = str(sd.nTSecurityDescriptor.Owner)
			acl.group_sid = str(sd.nTSecurityDescriptor.Group)
			acl.ace_order = order_ctr
			
			order_ctr += 1
			acl.guid = str(sd.objectGUID)
			if sd.objectSid:
				acl.sid = str(sd.objectSid)
			if sd.cn:
				acl.cn = sd.cn
			if sd.distinguishedName:
				acl.dn = str(sd.distinguishedName)
			acl.sd_control = sd.nTSecurityDescriptor.Control
			
			acl.ace_type = ace.Header.AceType.name
			acl.ace_mask = ace.Mask
			t = getattr(ace,'ObjectType', None)
			if t:
				acl.ace_objecttype = str(t)
			
			t = getattr(ace,'InheritedObjectType', None)
			if t:
				acl.ace_inheritedobjecttype = str(t)
				
			true_attr, false_attr = JackDawADDACL.mask2attr(ace.Mask)
			
			for attr in true_attr:	
				setattr(acl, attr, True)
			for attr in false_attr:	
				setattr(acl, attr, False)
				
			true_attr, false_attr = JackDawADDACL.hdrflag2attr(ace.Header.AceFlags)
			
			for attr in true_attr:	
				setattr(acl, attr, True)
			for attr in false_attr:	
				setattr(acl, attr, False)
			
			acl.ace_sid = str(ace.Sid)
			self.session.add(acl)
		
		self.session.commit()

	def check_status(self):
		#print('user')
		#print(self.user_ctr)
		#print(self.user_finish_ctr)
		#
		#print('machine')
		#print(self.machine_ctr)
		#print(self.machine_finish_ctr)
		#
		#print('ou')
		#print(self.ou_ctr)
		#print(self.ou_finish_ctr)
		#
		#print('group')
		#print(self.group_ctr)
		#print(self.group_finish_ctr)
		#
		#print('sd')
		#print(self.sd_ctr)
		#print(self.sd_finish_ctr)
		#
		#print('spn')
		#print(self.spn_ctr)
		#print(self.spn_finish_ctr)
		#
		#print('member')
		#print(self.member_ctr)
		#print(self.member_finish_ctr)
		#
		#print('domain')
		#print(self.domaininfo_ctr)
		#print(self.domaininfo_finish_ctr)

		if self.user_ctr == self.user_finish_ctr \
			and self.machine_ctr == self.machine_finish_ctr\
			and self.ou_ctr == self.ou_finish_ctr\
			and self.group_ctr == self.group_finish_ctr\
			and self.sd_ctr == self.sd_finish_ctr\
			and self.spn_ctr == self.spn_finish_ctr\
			and self.member_ctr == self.member_finish_ctr\
			and self.domaininfo_ctr == self.domaininfo_finish_ctr:
			return True
		return False

	def stop_agents(self):
		logger.info('mgr stop')
		self.session.commit()
		self.session.close()
		for _ in self.agents:
			self.agent_in_q.put(None)
		for agent in self.agents:
			agent.join()
		logger.info('stopped all agents!')

	def run(self):
		self.setup()
		logger.info('setup finished!')

		self.domaininfo_ctr += 1
		job = LDAPAgentJob(LDAPAgentCommand.DOMAININFO, None)
		self.agent_in_q.put(job)

		while True:
			res = self.agent_out_q.get()
			#print(res)
			res_type, res = res
			if res_type == LDAPAgentCommand.DOMAININFO:
				self.enum_domain(res)
			
			elif res_type == LDAPAgentCommand.USER:
				self.enum_user(res)

			elif res_type == LDAPAgentCommand.MACHINE:
				self.enum_machine(res)

			elif res_type == LDAPAgentCommand.SPNSERVICE:
				self.store_spnservice(res)

			elif res_type == LDAPAgentCommand.GROUP:
				self.enum_group(res)

			elif res_type == LDAPAgentCommand.OU:
				self.enum_ou(res)
			
			elif res_type == LDAPAgentCommand.SD:
				self.store_sd(res)
			
			elif res_type == LDAPAgentCommand.MEMBERSHIP:
				self.store_membership(res)

			elif res_type == LDAPAgentCommand.EXCEPTION:
				logger.warning(res)

			elif res_type == LDAPAgentCommand.SPNSERVICES_FINISHED:
				self.spn_finish_ctr += 1
			elif res_type == LDAPAgentCommand.USERS_FINISHED:
				self.user_finish_ctr += 1
			elif res_type == LDAPAgentCommand.MACHINES_FINISHED:
				self.machine_finish_ctr += 1
			elif res_type == LDAPAgentCommand.OUS_FINISHED:
				self.ou_finish_ctr += 1
			elif res_type == LDAPAgentCommand.GROUPS_FINISHED:
				self.group_finish_ctr += 1
			elif res_type == LDAPAgentCommand.MEMBERSHIPS_FINISHED:
				self.member_finish_ctr += 1
			elif res_type == LDAPAgentCommand.SDS_FINISHED:
				self.sd_finish_ctr += 1
			elif res_type == LDAPAgentCommand.DOMAININFO_FINISHED:
				self.domaininfo_finish_ctr += 1

			if self.check_status() == True:
				break
		
		self.stop_agents()
		return self.ad_id

		
if __name__ == '__main__':
	from msldap.commons.url import MSLDAPURLDecoder

	import sys
	sql = sys.argv[1]
	ldap_conn_url = sys.argv[2]

	print(sql)
	print(ldap_conn_url)
	
	ldap_mgr = MSLDAPURLDecoder(ldap_conn_url)

	mgr = LDAPEnumeratorManager(sql, ldap_mgr)
	mgr.run()


# System modules
from datetime import datetime
from jackdaw.nest.graph.domain import DomainGraph
from jackdaw.nest.graph.graphdata import GraphData
from jackdaw.nest.graph.construct import GraphConstruct
from jackdaw.nest.graph.domaindiff import DomainDiff

# 3rd party modules
from flask import make_response, abort, current_app
from flask import current_app


graph_id_ctr = 1
graphs = {}

diff_id_ctr = 1
diffs = {}

def list_all():
    return list(graphs.keys())

def create(adids):
	global graph_id_ctr
	db = current_app.db
	dg = DomainGraph(dbsession = db.session)
	for adid in adids:
		construct = GraphConstruct(adid)
		dg.construct(construct)
	graphs[graph_id_ctr] = dg
	graph_id_ctr += 1
	return {'graphid' : graph_id_ctr - 1}

def delete(graphid):
	return {}

def get(graphid):
	if graphid not in graphs:
		return 'Graph Not Found', 404
	res = graphs[graphid].all_shortest_paths()
	return res.to_dict()

def query_path(graphid, src = None, dst = None, format = 'd3'):
	if graphid not in graphs:
		return 'Graph Not Found', 404
	if src is None and dst is None:
		return {}
	res =  graphs[graphid].all_shortest_paths(src, dst)
	return res.to_dict(format = format)

def query_path_da(graphid, format = 'vis'):
	if graphid not in graphs:
		return 'Graph Not Found', 404
	
	da_sids = {}
	#searching for domain admin SID
	for node in graphs[graphid].get_node():
		if node.id.endswith('-512'):
			da_sids[node.id] = 1
	
	#print(da_sids)
	if len(da_sids) == 0:
		return 'No domain administrator group found', 404
	
	res = GraphData()
	for sid in da_sids:
		res += graphs[graphid].all_shortest_paths(None, sid)


	#print(res)
	return res.to_dict(format = format)


def list_nodes(graphid, with_data = False):
	if graphid not in graphs:
		return 'Graph Not Found', 404
	nodes = []
	for node in graphs[graphid].get_node():
		nodes.append(node.to_dict())
	return nodes

def get_node(graphid, nodeid):
	if graphid not in graphs:
		return 'Graph Not Found', 404
	return graphs[graphid].get_node(nodeid)

def query_path_all(graphid):
	if graphid not in graphs:
		return 'Graph Not Found', 404
	return graphs[graphid].show_all().to_dict(format = 'vis')

def search_sid(graphid, sid):
	return {}

def search_cn(graphid, cn):
	return {}

def stat_distance(graphid, sid):
	if graphid not in graphs:
		return 'Graph Not Found', 404
	distances = graphs[graphid].distances_from_node(sid)
	
	return distances

def diff(graphids):
	global diff_id_ctr
	
	db = current_app.db
	dd = DomainDiff(dbsession=db.session)
	construct_old = GraphConstruct(graphids['src'])
	construct_new = GraphConstruct(graphids['dst'])

	diffs[diff_id_ctr] = dd
	diff_id_ctr += 1

	dd.construct(construct_old, construct_new)

	return {
		'diffid' : diff_id_ctr - 1
	}

def list_diff_all():
	return list(diffs.keys())

def diff_nodes(diffid):
	diffres = diffs[diffid].diff_nodes()
	return diffres

def diff_path_distance(diffid, sid):
	diffres = diffs[diffid].diff_path_distance(sid)
	return diffres

def diff_path(diffid, srcsid, dstsid):
	diffres = diffs[diffid].diff_path(srcsid, dstsid)
	return diffres

def diff_path_da(diffid):
	diffres = diffs[diffid].diff_path_da()
	return diffres



from flask import current_app
from jackdaw.dbmodel.aduser import JackDawADUser

def list_users(domainid, page, maxcnt):
	db = current_app.db
	res = {
		'res' : [],
		'page': {},
	}
	qry = db.session.query(
        JackDawADUser
        ).filter_by(ad_id = domainid
        ).with_entities(
            JackDawADUser.id, 
            JackDawADUser.objectSid, 
            JackDawADUser.sAMAccountName
            )
        
	
	qry = qry.paginate(page = page, max_per_page = maxcnt)

	domains = []
	for uid, sid, name in qry.items:
		domains.append(
			{
				'id' : uid, 
				'sid' : sid, 
				'name': name
			}
		)

	page = dict(
		total=qry.total, 
		current_page=qry.page,
		per_page=qry.per_page
	)

	res['res'] = domains
	res['page'] = page

	return res

def get(domainid, userid):
    db = current_app.db
    user = db.session.query(JackDawADUser).get(userid)
    return user.to_dict()

def get_sid(domainid, usersid):
    db = current_app.db
    for user in db.session.query(JackDawADUser).filter_by(objectSid = usersid).filter(JackDawADUser.ad_id == domainid).all():
        return user.to_dict()

def filter(domainid, proplist):
    #TODO: add other properties to search for!
    db = current_app.db
    query = db.session.query(JackDawADUser).filter_by(ad_id = domainid)
    for elem in proplist:
        if 'sAMAccountName' in elem:
            query = query.filter(JackDawADUser.sAMAccountName == elem['sAMAccountName'])
    
    user = query.first()
    if user is None:
        return {}
    return user.to_dict()


from . import Basemodel, lf
import datetime
import hashlib
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, ForeignKey, Index

from pypykatz.pypykatz import pypykatz
from pypykatz.utils.crypto.winhash import LM, NT

# It may be tempting to use SIDs to link credentials with users in the domain
# However some credentials format don't give SIDs (impacket) others have SIDs 
# that only identify the primary user, but not the owner of the actual credential
# Summary: we need to use the ad_id and the username<->samaccountname
#

class Credential(Basemodel):
	__tablename__ = 'credentials'
	__table_args__ = (Index('Credential_uc', "ad_id", "domain", "username","nt_hash", "lm_hash", "history_no", unique = True), )

	id = Column(Integer, primary_key=True)
	ad_id = Column(Integer)
	domain = Column(String, index=True, nullable= False)
	username = Column(String, index=True, nullable= False)
	nt_hash = Column(String, index=True, nullable= False)
	lm_hash = Column(String, index=True, nullable= False)
	history_no = Column(Integer, index=True, nullable= False)
	cred_type = Column(String, index=True, nullable= False)
	
	def __init__(self, domain = None, username = None, nt_hash = None, lm_hash = None, history_no = None, ad_id = -1):
		self.domain = domain
		self.username = username
		self.nt_hash = nt_hash
		self.lm_hash = lm_hash
		self.history_no = history_no
		self.ad_id = ad_id

	@staticmethod
	def from_impacket_line(line, ad_id = -1):
		cred = Credential()
		userdomainhist, flags, lm_hash, nt_hash, *t = line.split(':')
		#parsing history
		m = userdomainhist.find('_history')
		history_no = 0
		if m != -1:
			history_no = int(userdomainhist.split('_history')[1]) + 1
			userdomainhist = userdomainhist.split('_history')[0]
		m = userdomainhist.find('\\')
		domain = '<LOCAL>'
		username = userdomainhist
		if m != -1:
			domain = userdomainhist.split('\\')[0]
			username = userdomainhist.split('\\')[1]
		cred.ad_id = ad_id
		cred.nt_hash = nt_hash
		cred.lm_hash = lm_hash
		cred.history_no = history_no
		cred.username = username
		cred.domain = domain
		cred.cred_type = 'dcsync'
		return cred

	@staticmethod
	def from_impacket_stream(stream, ad_id = -1):
		for line in stream:
			yield Credential.from_impacket_line(line.decode(), ad_id = ad_id)

	@staticmethod
	def from_impacket_file(filename, ad_id = -1):
		"""
		Remember that this doesnt populate the foreign keys!!! You'll have to do it separately!
		important: historyno will start at 0. This means all history numbers in the file will be incremented by one
		"""
		with open(filename, 'r') as f:
			for line in f:
				yield Credential.from_impacket_line(line, ad_id = ad_id)

	@staticmethod
	def from_lsass_stream(stream, ad_id = -1):
		from pypykatz.pypykatz import pypykatz
		mimi = pypykatz.parse_minidump_buffer(stream)
		return Credential.lsass_generator(mimi, ad_id = ad_id)

	@staticmethod
	def from_lsass_dump(filename, ad_id = -1):
		
		mimi = pypykatz.parse_minidump_file(filename)
		return Credential.lsass_generator(mimi, ad_id = ad_id)

	@staticmethod
	def lsass_generator(mimi, ad_id):

		for luid in mimi.logon_sessions:
			sid = mimi.logon_sessions[luid].sid

			for cred in mimi.logon_sessions[luid].msv_creds:
				cr = Credential()
				cr.ad_id = ad_id
				cr.nt_hash = cred.NThash.hex() if cred.NThash is not None else '31d6cfe0d16ae931b73c59d7e0c089c0'
				cr.lm_hash = cred.LMHash if cred.LMHash is not None else 'aad3b435b51404eeaad3b435b51404ee'
				cr.history_no = 0
				cr.username = cred.username if cred.username is not None else 'NA'
				cr.domain = cred.domainname if cred.domainname is not None else '<LOCAL>'
				cr.cred_type = 'msv'
				yield cr, None, sid

			for cred in mimi.logon_sessions[luid].wdigest_creds:
				if cred.password is not None:
					cr = Credential()
					cr.ad_id = ad_id
					cr.nt_hash = NT(cred.password).hex()
					cr.lm_hash = None
					cr.history_no = 0
					cr.username = cred.username if cred.username is not None else 'NA'
					cr.domain = cred.domainname if cred.domainname is not None else '<LOCAL>'
					cr.cred_type = 'wdigest'
					yield cr, cred.password, sid

			for cred in mimi.logon_sessions[luid].ssp_creds:
				if cred.password is not None:
					cr = Credential()
					cr.ad_id = ad_id
					cr.nt_hash = NT(cred.password).hex()
					cr.lm_hash = None
					cr.history_no = 0
					cr.username = cred.username if cred.username is not None else 'NA'
					cr.domain = cred.domainname if cred.domainname is not None else '<LOCAL>'
					cr.cred_type = 'ssp'
					yield cr, cred.password, sid

			for cred in mimi.logon_sessions[luid].livessp_creds:
				if cred.password is not None:
					cr = Credential()
					cr.ad_id = ad_id
					cr.nt_hash = NT(cred.password).hex()
					cr.lm_hash = None
					cr.history_no = 0
					cr.username = cred.username if cred.username is not None else 'NA'
					cr.domain = cred.domainname if cred.domainname is not None else '<LOCAL>'
					cr.cred_type = 'live_ssp'
					yield cr, cred.password, sid

			#for cred in mimi.logon_sessions[luid]['dpapi_creds']:
			# dpapi credentials are not used in this database (for now)

			for cred in mimi.logon_sessions[luid].kerberos_creds:
				if cred.password is not None:
					cr = Credential()
					cr.ad_id = ad_id
					cr.nt_hash = NT(cred.password).hex()
					cr.lm_hash = None
					cr.history_no = 0
					cr.username = cred.username if cred.username is not None else 'NA'
					cr.domain = cred.domainname if cred.domainname is not None else '<LOCAL>'
					cr.cred_type = 'kerberos'
					yield cr, cred.password, sid

			for cred in mimi.logon_sessions[luid].credman_creds:
				if cred.password is not None:
					cr = Credential()
					cr.ad_id = ad_id
					cr.nt_hash = NT(cred.password).hex()
					cr.lm_hash = None
					cr.history_no = 0
					cr.username = cred.username if cred.username is not None else 'NA'
					cr.domain = cred.domainname if cred.domainname is not None else '<LOCAL>'
					cr.cred_type = 'credman'
					yield cr, cred.password, sid

			for cred in mimi.logon_sessions[luid].tspkg_creds:
				if cred.password is not None:
					cr = Credential()
					cr.ad_id = ad_id
					cr.nt_hash = NT(cred.password).hex()
					cr.lm_hash = None
					cr.history_no = 0
					cr.username = cred.username if cred.username is not None else 'NA'
					cr.domain = cred.domainname if cred.domainname is not None else '<LOCAL>'
					cr.cred_type = 'tpskg'
					yield cr, cred.password, sid

				
	@staticmethod
	def from_aiosmb_stream(stream, ad_id = -1):
		for line in stream:
			if line is None or len(line) == 0:
				continue
			yield Credential.from_aiosmb_line(line.decode(), ad_id = ad_id)

	@staticmethod
	def from_aiosmb_file(filename, ad_id = -1):
		"""
		Remember that this doesnt populate the foreign keys!!! You'll have to do it separately!
		important: historyno will start at 0. This means all history numbers in the file will be incremented by one
		"""
		with open(filename, 'r') as f:
			for line in f:
				if line is None or len(line) == 0:
					continue
				yield Credential.from_aiosmb_line(line, ad_id = ad_id)


	@staticmethod
	def from_aiosmb_line(line, ad_id = -1):
		cred = Credential()
		line = line.strip()
		data = line.split(':')
		pw = None
		if data[0] == 'ntlm':
			cred.ad_id = ad_id
			cred.domain = data[1]
			cred.username = data[2]
			cred.nt_hash = data[5]
			cred.lm_hash = data[4]
			cred.history_no = 0
			cred.cred_type = 'aiosmb-dcsync-ntlm'

		elif data[0] == 'ntlm_history':
			cred.ad_id = ad_id
			cred.domain = data[1]
			cred.username = data[2]
			cred.nt_hash = data[5]
			cred.lm_hash = data[4]
			cred.history_no = int(data[6].replace('history_',''))
			cred.cred_type = 'aiosmb-dcsync-ntlm-history'

		elif data[0] == 'cleartext':
			cred.ad_id = ad_id
			_, cred.domain, cred.username, pw = line.split(':', 4) #reparsing needed, pw might contain colon
			
			cred.nt_hash = NT(pw).hex()
			cred.lm_hash = None
			cred.history_no = 0
			cred.cred_type = 'aiosmb-dcsync-cleartext'

		return cred, pw


from flask import current_app
from jackdaw.dbmodel.adou import JackDawADOU

def list_ous(domainid, page, maxcnt):
	db = current_app.db
	res = {
		'res' : [],
		'page': {},
	}
	qry = db.session.query(
		JackDawADOU
		).filter_by(ad_id = domainid
		).with_entities(
			JackDawADOU.id, 
			JackDawADOU.objectGUID, 
			JackDawADOU.ou
			)
				
	qry = qry.paginate(page = page, max_per_page = maxcnt)

	domains = []
	for uid, sid, name in qry.items:
		domains.append(
			{
				'id' : uid, 
				'sid' : sid, 
				'name': name
			}
		)

	page = dict(
		total=qry.total, 
		current_page=qry.page,
		per_page=qry.per_page
	)

	res['res'] = domains
	res['page'] = page

	return res

def get(domainid, ouid):
	db = current_app.db
	user = db.session.query(JackDawADOU).get(ouid)
	return user.to_dict()

def get_guid(domainid, guid):
	db = current_app.db
	for user in db.session.query(JackDawADOU).filter_by(objectGUID = guid).filter(JackDawADOU.ad_id == domainid).all():
		return user.to_dict()